/* This file is part of Om.  Copyright (C) 2006 Dave Robillard.
 * 
 * Om is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * Om is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef CONTROLINTERFACE_H
#define CONTROLINTERFACE_H

#include <cassert>
#include <sigc++/sigc++.h>
#include <string>
using std::string;

namespace LibOmClient {
	class PluginModel;
	class PatchModel;
	class NodeModel;
	class PortModel;
	class ConnectionModel;
}
using namespace LibOmClient;

namespace OmGtk {

class App;


// Some boilerplate killing macros.. see their use below, should be obvious

#define SLOT_INIT(name) name##_slot(sigc::mem_fun(this, &ControlInterface:: name))

/** The downside of this macro is you must pass only the type names, since it's
 * used in the template parameter as well as the function parameter.  Doxygen
 * will still pull the functions declarations nicely though... */
#define SLOT(name, ...) \
sigc::slot<void, __VA_ARGS__> name##_slot; \
void name (__VA_ARGS__);



/** Provides the public interface for controlling OmGtk (via engine events).
 *
 * All control of OmGtk from the engine happens through this interface.
 *
 * All of these functions must be called in the GTK thread.  This is a unified
 * interface for controlling OmGtk (eg from the engine) but it doesn't take
 * care of any threading issues.  See GtkClientHooks for that.
 */
class ControlInterface : public sigc::trackable {
public:

	ControlInterface(App* app)
	: _app(app)
	, SLOT_INIT(error)
	, SLOT_INIT(new_plugin)
	, SLOT_INIT(new_patch)
	, SLOT_INIT(new_node)
	, SLOT_INIT(new_port)
	, SLOT_INIT(patch_enabled)
	, SLOT_INIT(patch_disabled)
	, SLOT_INIT(patch_cleared)
	, SLOT_INIT(object_destroyed)
	, SLOT_INIT(object_renamed)
	, SLOT_INIT(connection)
	, SLOT_INIT(disconnection)
	, SLOT_INIT(metadata_update)
	, SLOT_INIT(control_change)
	, SLOT_INIT(program_add)
	, SLOT_INIT(program_remove)
	{
		assert(_app);
	}

	SLOT(error, const string);
	SLOT(new_plugin, PluginModel*);
	SLOT(new_patch, PatchModel*);
	SLOT(new_node, NodeModel*);
	SLOT(new_port, PortModel*);
	SLOT(patch_enabled, const string);
	SLOT(patch_disabled, const string);
	SLOT(patch_cleared, const string);
	SLOT(object_destroyed, const string);
	SLOT(object_renamed, const string, const string);
	SLOT(connection, ConnectionModel*);
	SLOT(disconnection, const string, const string);
	SLOT(metadata_update, const string, const string, const string);
	SLOT(control_change, const string, float);
	SLOT(program_add, const string, int, int, const string);
	SLOT(program_remove, const string, int, int);

private:
	App* _app;
};


} // namespace OmGtk

#endif // CONTROLINTERFACE_H

