/* This file is part of Om.  Copyright (C) 2006 Dave Robillard.
 * 
 * Om is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * Om is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef CONTROLLER_H
#define CONTROLLER_H

#include <cassert>
#include <string>
#include <list>
#include <glibmm.h>
#include "PluginModel.h"

namespace LibOmClient {
class PatchModel;
class NodeModel;
class PatchLibrarian;
class OSCController;
}

using std::string; using std::list;
using namespace LibOmClient;

namespace OmGtk {

class PatchController;
class GtkClientHooks;
class Loader;
class ControlInterface;

	
/** Singleton engine controller for the entire app.
 *
 * This is hardly more than a trivial wrapper for OSCController, suggesting something
 * needs a rethink around here..
 *
 * This needs to be either eliminated or the name changed to OmController.  Probably
 * best to keep around in case multiple engine control support comes around?
 *
 * \ingroup OmGtk
 */
class Controller
{
public:
	~Controller();

	void attach();
	
	void register_client();
	void load_plugins();

	string engine_url() const;
	void   engine_url(const string& url);

	int  client_port() const;
	void client_port(int port);
	
	size_t num_plugins() const;

	void activate_engine();
	void deactivate_engine();
	void enable_engine();
	void disable_engine();

	void quit();

	void set_control(const string& port_path, float val);
	void set_control(const string& port_path, int voice, float val);
	
	void set_control_slow(const string& port_path, float val);

	void set_program(const string& node_path, int bank, int program);
	
	void request_control(const string& port_path);

	void add_node(NodeModel* nm);

	void connect(const string& port1_path, const string& port2_path);
	void disconnect(const string& port1_path, const string& port2_path);
	void disconnect_all(const string& node_path);
	
	void load_patch(PatchModel* model, bool wait = true, bool merge = false);
	void save_patch(PatchModel* model, const string& filename, bool recursive);

	void create_patch(PatchModel* pm);
	void clear_patch(const string& path);

	void destroy(const string& path);
	
	void rename(const string& path, const string& name);
	
	void enable_patch(const string& patch_path);
	void disable_patch(const string& patch_path);
	
	void midi_learn(const string& node_path);

	void set_metadata(const string& object_path, const string& key, const string& value);
	
	void lash_restore_finished();

	void request_plugins();
	void request_all_objects();
	
	bool is_attached() const;

	void set_patch_path(const string& path);

	/*
	void push_added_node(NodeModel* nm) { m_added_nodes.push_back(nm); }
	NodeModel* yank_added_node(const string& path);

	void push_added_patch(PatchModel* pm) { m_added_patches.push_back(pm); }
	PatchModel* yank_added_patch(const string& path);
	*/
	GtkClientHooks*   client_hooks()      { return m_client_hooks; }

	static void instantiate(ControlInterface* interface)
	{ if (!m_controller) m_controller = new Controller(interface); }

	inline static Controller& instance() { assert(m_controller); return *m_controller; }

private:
	Controller(ControlInterface* interface);
	static Controller* m_controller;
	
	GtkClientHooks*   m_client_hooks;
	OSCController*    m_osc_controller;
	PatchLibrarian*   m_patch_librarian;
	Loader*           m_loader;

	/** Used to cache added nodes so client can remember some values when the
	 * new node notification comes (location, etc).  Used to prevent the node
	 * jumping locations during the delay between new node and the module-x/y
	 * metadata notifications */
	//list<NodeModel*> m_added_nodes;

	/** Used to cache added patches in the same was as m_added_nodes.  Used to
	 * rember filename so File->Save can work without prompting (filename can't
	 * be sent to the server as metadata, because a client on another machine
	 * doesn't want that information) */
	//list<PatchModel*> m_added_patches;
	
	string m_engine_url;
	int    m_client_port; ///< Set by main() before attaching
};


} // namespace OmGtk

#endif // CONTROLLER_H
