/* This file is part of Om.  Copyright (C) 2006 Dave Robillard.
 * 
 * Om is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * Om is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef GTKCLIENTHOOKS_H
#define GTKCLIENTHOOKS_H

#include <string>
#include <queue>
#include <iostream>
#include <sigc++/sigc++.h>
#include "ClientHooks.h"
#include "ControlInterface.h"
#include "Queue.h"
using std::string;
using std::cerr; using std::endl;

namespace LibOmClient { class PluginModel; }
using namespace LibOmClient;

namespace OmGtk {

/** Returns nothing and takes no parameters (because they have been bound) */
typedef sigc::slot<void> Closure;


/** ClientHooks implementation for the Gtk client.
 *
 * This is a threadsafe interface to OmGtk.  It provides the same interface as
 * @ref ControlInterface, except all public functions may be called in a thread
 * other than the GTK thread (a closure will be created and pushed through a
 * queue for the GTK thread to execute).
 *
 * \ingroup OmGtk
 */
class GtkClientHooks : public ClientHooks
{
public:
	GtkClientHooks(ControlInterface* interface);
	
	void set_ignore_port(const string& path) { m_ignore_port_path = path; }
	void clear_ignore_port()                 { m_ignore_port_path = ""; }
	
	// OSC thread functions (deferred calls)
	
	void error(const string& msg)
		{ push_event(sigc::bind(_interface->error_slot, msg)); }
	
	void new_plugin(PluginModel* const pm)
		{ push_event(sigc::bind(_interface->new_plugin_slot, pm)); }
	
	void new_patch(PatchModel* const pm)
		{ push_event(sigc::bind(_interface->new_patch_slot, pm)); }

	void new_node(NodeModel* const nm)
		{ push_event(sigc::bind(_interface->new_node_slot, nm)); }

	void new_port(PortModel* const pm)
		{ push_event(sigc::bind(_interface->new_port_slot, pm)); }
	
	void destroyed(const string& path)
		{ push_event(sigc::bind(_interface->object_destroyed_slot, path)); }
	
	void patch_enabled(const string& path)
		{ push_event(sigc::bind(_interface->patch_enabled_slot, path)); }
	
	void patch_disabled(const string& path)
		{ push_event(sigc::bind(_interface->patch_disabled_slot, path)); }

	void patch_cleared(const string& path)
		{ push_event(sigc::bind(_interface->patch_cleared_slot, path)); }

	void object_renamed(const string& old_path, const string& new_path)
		{ push_event(sigc::bind(_interface->object_renamed_slot, old_path, new_path)); }
	
	void connection(ConnectionModel* const cm)
		{ push_event(sigc::bind(_interface->connection_slot, cm)); }
	
	void disconnection(const string& src_port_path, const string& dst_port_path)
		{ push_event(sigc::bind(_interface->disconnection_slot, src_port_path, dst_port_path)); }
	
	void metadata_update(const string& path, const string& key, const string& value)
		{ push_event(sigc::bind(_interface->metadata_update_slot, path, key, value)); }

	void control_change(const string& port_path, float value)
		{ push_event(sigc::bind(_interface->control_change_slot, port_path, value)); }

	void program_add(const string& path, int bank, int program, const string& name)
		{ push_event(sigc::bind(_interface->program_add_slot, path, bank, program, name)); }
	
	void program_remove(const string& path, int bank, int program)
		{ push_event(sigc::bind(_interface->program_remove_slot, path, bank, program)); }

	/** Process all queued events - MUST be called from Gtk thread.
	 * Registered as a GTK idle handler by App. */
	bool process_events();

private:
	
	/** Provides the functions/slots the closures will actually call in the GTK thread */
	ControlInterface* _interface;

	void push_event(Closure ev);

	/** Set if a port slider is grabbed and is being dragged.
	 * If a control event comes in for a port with this path, we'll just
	 * ignore it outright.  (Just an optimization over doing all the searching
	 * for the port slider just to ignore the event) */
	string m_ignore_port_path;

	Queue<Closure> _events;
};


} // namespace OmGtk

#endif // GTKCLIENTHOOKS_H
