/* This file is part of Om.  Copyright (C) 2005 Dave Robillard.
 * 
 * Om is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * Om is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "OSCReceiver.h"
#include <iostream>
#include <cstdlib>
#include <string>
#include <lo/lo.h>
#include "types.h"
#include "Om.h"
#include "OmApp.h"
#include "Queue.h"
#include "PreProcessor.h"
#include "OSCSender.h"
#include "OSCResponder.h"
#include "ActivateEvent.h"
#include "DeactivateEvent.h"
#include "EnablePatchEvent.h"
#include "DisablePatchEvent.h"
#include "ClearPatchEvent.h"
#include "SetPortValueEvent.h"
#include "SetPortValueSlowEvent.h"
#include "ConnectionEvent.h"
#include "DisconnectionEvent.h"
#include "AddNodeEvent.h"
#include "CreatePatchEvent.h"
#include "DestroyEvent.h"
#include "SetMetadataEvent.h"
#include "RequestMetadataEvent.h"
#include "RequestPortValueEvent.h"
#include "RequestAllObjectsEvent.h"
#include "RequestPluginsEvent.h"
#include "LoadPluginsEvent.h"
#include "NoteOnEvent.h"
#include "NoteOffEvent.h"
#include "AllNotesOffEvent.h"
#include "DisconnectNodeEvent.h"
#include "RegisterClientEvent.h"
#include "UnregisterClientEvent.h"
#include "RenameEvent.h"
#include "PingSlowEvent.h"
#include "MidiLearnEvent.h"
#include "Plugin.h"

#ifdef HAVE_LASH
#include "LashRestoreDoneEvent.h"
#endif

#ifdef HAVE_DSSI
#include "DSSIUpdateEvent.h"
#include "DSSIControlEvent.h"
#include "DSSIConfigureEvent.h"
#include "DSSIProgramEvent.h"
#endif

using std::cerr; using std::cout; using std::endl;

namespace Om {

/*! \page engine_osc_namespace Engine OSC Namespace Documentation
 *
 * <p>These are the commands the engine recognizes.  A client can control every
 * aspect of the engine entirely with these commands.</p>
 *
 * <p>All commands on this page are in the "control band".  If a client needs to
 * know about the state of the engine, it must listen to the "notification band".
 * See the "Client OSC Namespace Documentation" for details.
 */


OSCReceiver::OSCReceiver(const char* const port)
: m_port(port),
  m_is_activated(false),
  m_st(NULL)
{
	m_st = lo_server_thread_new(port, error_cb);
	
	if (m_st == NULL) {
		cerr << "[OSC] Could not start OSC server.  Aborting." << endl;
		exit(EXIT_FAILURE);
	} else {
		char* lo_url = lo_server_thread_get_url(m_st);
		cout << "[OSC] Started OSC server at " << lo_url << endl;
		free(lo_url);
	}

	// For debugging, print all incoming OSC messages
	//lo_server_thread_add_method(m_st, NULL, NULL, generic_cb, NULL);

	// Commands
	lo_server_thread_add_method(m_st, "/om/ping", "i", ping_cb, this);
	lo_server_thread_add_method(m_st, "/om/ping_slow", "i", ping_slow_cb, this);
	lo_server_thread_add_method(m_st, "/om/engine/quit", "i", quit_cb, this);
	lo_server_thread_add_method(m_st, "/om/engine/register_client", "is", register_client_cb, this);
	lo_server_thread_add_method(m_st, "/om/engine/register_client", "i", register_client_cb, this);
	lo_server_thread_add_method(m_st, "/om/engine/unregister_client", "i", unregister_client_cb, this);
	lo_server_thread_add_method(m_st, "/om/engine/load_plugins", "i", load_plugins_cb, this);
	lo_server_thread_add_method(m_st, "/om/engine/activate", "i", engine_activate_cb, this);
	lo_server_thread_add_method(m_st, "/om/engine/deactivate", "i", engine_deactivate_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/create_patch", "isi", create_patch_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/enable_patch", "is", enable_patch_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/disable_patch", "is", disable_patch_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/clear_patch", "is", clear_patch_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/create_node", "issssi", create_node_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/create_node", "isssi", create_node_by_uri_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/destroy", "is", destroy_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/rename", "iss", rename_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/connect", "iss", connect_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/disconnect", "iss", disconnect_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/disconnect_all", "is", disconnect_all_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/set_port_value", "isf", set_port_value_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/set_port_value", "isif", set_port_value_voice_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/set_port_value_slow", "isf", set_port_value_slow_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/note_on", "isii", note_on_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/note_off", "isi", note_off_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/all_notes_off", "isi", all_notes_off_cb, this);
	lo_server_thread_add_method(m_st, "/om/synth/midi_learn", "is", midi_learn_cb, this);
#ifdef HAVE_LASH
	lo_server_thread_add_method(m_st, "/om/lash/restore_finished", "i", lash_restore_done_cb, this);
#endif
	
	// DSSI support
#ifdef HAVE_DSSI
	lo_server_thread_add_method(m_st, NULL, NULL, dssi_cb, this);
#endif

	lo_server_thread_add_method(m_st, "/om/metadata/request", "isss", metadata_get_cb, this);
	lo_server_thread_add_method(m_st, "/om/metadata/set", "isss", metadata_set_cb, this);
	
	// Queries
	lo_server_thread_add_method(m_st, "/om/request/plugins", "i", request_plugins_cb, this);
	lo_server_thread_add_method(m_st, "/om/request/all_objects", "i", request_all_objects_cb, this);
	lo_server_thread_add_method(m_st, "/om/request/port_value", "is", request_port_value_cb, this);
	
	lo_server_thread_add_method(m_st, NULL, NULL, unknown_cb, NULL);
}


OSCReceiver::~OSCReceiver()
{
	deactivate();

	if (m_st != NULL)  {
		lo_server_thread_free(m_st);
		m_st = NULL;
	}
}


void
OSCReceiver::activate()
{
	if (!m_is_activated) {
		lo_server_thread_start(m_st);
		m_is_activated = true;
	}
	
	/* Waiting on the next liblo release
	pthread_t lo_thread = lo_server_thread_get_thread(m_st);
	
	sched_param sp;
	sp.sched_priority = 20;
	int result = pthread_setschedparam(lo_thread, SCHED_FIFO, &sp);
	if (!result)
		cout << "[OSC] Set OSC thread to realtime scheduling (SCHED_FIFO, priority "
			<< sp.sched_priority << ")" << endl;
	else
		cout << "[OSC] Unable to set OSC thread to realtime scheduling ("
			<< strerror(result) << endl;
	*/
}


void
OSCReceiver::deactivate()
{
	if (m_is_activated) {
		lo_server_thread_stop(m_st);
		cout << "[OSCReceiver] Stopped OSC server thread" << endl;
		m_is_activated = false;
	}
}


void
OSCReceiver::error_cb(int num, const char* msg, const char* path)
{
	cerr << "liblo server error " << num << " in path \"" << "\" - " << msg << endl;
}


/** \page engine_osc_namespace
 * <p> \b /om/ping - Immediately sends a successful response to the given response id.
 * \arg \b response-id (integer) </p> \n \n
 */
int
OSCReceiver::m_ping_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address addr       = lo_message_get_source(msg);
	const int  response_id = argv[0]->i;
	
	OSCResponder(addr, response_id).respond_ok();
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/ping_slow - Sends response after going through the ("slow") event queue.
 * \arg \b response-id (integer)
 *
 * \li See the documentation for /om/synth/set_port_value_slow for an explanation of how
 * this differs from /om/ping.  This is useful to send after sending a large cluster of
 * events as a sentinel and wait on it's response, to know when the events are all
 * finished processing.</p> \n \n
 */
int
OSCReceiver::m_ping_slow_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address addr       = lo_message_get_source(msg);
	const int  response_id = argv[0]->i;
	
	PingSlowEvent* ev = new PingSlowEvent(new OSCResponder(addr, response_id));
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/engine/quit - Terminates the engine.
 * \arg \b response-id (integer)
 *
 * \li Note that there is NO order guarantees with this command at all.  You could
 * send 10 messages then quit, and the quit reply could come immediately and the
 * 10 messages would never get executed.  This is the client's problem, for the
 * time being.  Complain if you want it fixed. </p> \n \n
 */
int
OSCReceiver::m_quit_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address addr        = lo_message_get_source(msg);
	const int  response_id = argv[0]->i;

	OSCResponder(addr, response_id).respond_ok();
	
	om->quit();

	return 0;  // handled
}


/** \page engine_osc_namespace
 * <p> \b /om/engine/register_client - Registers a new client with the engine
 * \arg \b response-id (integer)
 * \arg \b url (string) - URL to send notifications to (if not present, use source address) </p> \n \n
 */
int
OSCReceiver::m_register_client_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address        addr         = lo_message_get_source(msg);
	const int         response_id  =  argv[0]->i;
	const char* const url          = (argc == 1) ? lo_address_get_url(addr) : &argv[1]->s;

	RegisterClientEvent* ev = new RegisterClientEvent(new OSCResponder(addr, response_id), url);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/engine/unregister_client - Unregisters a client
 * \arg \b response-id (integer) </p> \n \n
 */
int
OSCReceiver::m_unregister_client_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	
	UnregisterClientEvent* ev = new UnregisterClientEvent(new OSCResponder(addr, response_id));
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/engine/load_plugins - Locates all available plugins, making them available for use.
 * \arg \b response-id (integer) </p> \n \n
 */
int
OSCReceiver::m_load_plugins_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address addr        = lo_message_get_source(msg);
	const int  response_id = argv[0]->i;

	LoadPluginsEvent* ev = new LoadPluginsEvent(new OSCResponder(addr, response_id));
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/engine/activate - Activate the engine (MIDI, audio, everything)
 * \arg \b response-id (integer) </p>
 *
 * \li Note that you <b>must</b> send this message first if you want the engine to do
 * anything at all - <em>including respond to your messages!</em> \n \n
 */
int
OSCReceiver::m_engine_activate_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address addr        = lo_message_get_source(msg);
	const int  response_id = argv[0]->i;

	ActivateEvent* ev = new ActivateEvent(new OSCResponder(addr, response_id));
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/engine/deactivate - Deactivate the engine completely.
 * \arg \b response-id (integer) </p> \n \n
 */
int
OSCReceiver::m_engine_deactivate_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address addr        = lo_message_get_source(msg);
	const int  response_id = argv[0]->i;

	DeactivateEvent* ev = new DeactivateEvent(new OSCResponder(addr, response_id));
	om->pre_processor()->push(ev);

	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/create_patch - Creates a new, empty, toplevel patch.
 * \arg \b response-id  (integer)
 * \arg \b patch-path  (string) - Patch path (complete, ie /master/parent/new_patch)
 * \arg \b poly        (integer) - Patch's (internal) polyphony </p> \n \n
 */
int
OSCReceiver::m_create_patch_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* patch_path  = &argv[1]->s;
	const int   poly        =  argv[2]->i;

	CreatePatchEvent* ev = new CreatePatchEvent(new OSCResponder(addr, response_id), patch_path, poly);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/rename - Rename an Object (only Nodes, for now)
 * \arg \b response-id (integer)
 * \arg \b path - Object's path
 * \arg \b name - New name for object </p> \n \n
 */
int
OSCReceiver::m_rename_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr         = lo_message_get_source(msg);
	const int   response_id  =  argv[0]->i;
	const char* object_path  = &argv[1]->s;
	const char* name         = &argv[2]->s;
	
	RenameEvent* ev = new RenameEvent(new OSCResponder(addr, response_id), object_path, name);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/enable_patch - Enable DSP processing of a patch
 * \arg \b response-id (integer)
 * \arg \b patch-path - Patch's path
 */
int
OSCReceiver::m_enable_patch_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* patch_path  = &argv[1]->s;
	
	EnablePatchEvent* ev = new EnablePatchEvent(new OSCResponder(addr, response_id), patch_path);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/disable_patch - Disable DSP processing of a patch
 * \arg \b response-id (integer)
 * \arg \b patch-path - Patch's path
 */
int
OSCReceiver::m_disable_patch_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* patch_path  = &argv[1]->s;
	
	DisablePatchEvent* ev = new DisablePatchEvent(new OSCResponder(addr, response_id), patch_path);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/clear_patch - Remove all nodes from a patch
 * \arg \b response-id (integer)
 * \arg \b patch-path - Patch's path
 */
int
OSCReceiver::m_clear_patch_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* patch_path  = &argv[1]->s;
	
	ClearPatchEvent* ev = new ClearPatchEvent(new OSCResponder(addr, response_id), patch_path);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/create_node - Add a node into a given patch (load a plugin by URI)
 * \arg \b response-id (integer)
 * \arg \b node-path (string) - Full path of the new node (ie. /patch2/subpatch/newnode)
 * \arg \b type (string) - Plugin type ("Internal", "LV2", "DSSI", "LADSPA")
 * \arg \b plug-uri (string) - URI of the plugin to load
 * \arg \b poly (integer-boolean) - Whether node is polyphonic (0 = false, 1 = true) </p> \n \n
 */
int
OSCReceiver::m_create_node_by_uri_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* node_path   = &argv[1]->s;
	const char* type        = &argv[2]->s;
	const char* plug_uri    = &argv[3]->s;
	const int   poly        =  argv[4]->i;
	
	// FIXME Event-ize
	
	Plugin* plugin = new Plugin();
	plugin->set_type(type);
	plugin->uri(plug_uri);

	if (poly != 0 && poly != 1) {
		OSCResponder(addr, response_id).respond_error("Invalid poly parameter in create_node");
		return 0;
	}

	AddNodeEvent* ev = new AddNodeEvent(new OSCResponder(addr, response_id),
		node_path, plugin, (poly == 1));
	om->pre_processor()->push(ev);

	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/create_node - Add a node into a given patch (load a plugin by libname, label)
 * \arg \b response-id (integer)
 * \arg \b node-path (string) - Full path of the new node (ie. /patch2/subpatch/newnode)
 * \arg \b type (string) - Plugin type ("LADSPA" or "Internal")
 * \arg \b lib-name (string) - Name of library where plugin resides (eg "cmt.so")
 * \arg \b plug-label (string) - Label (ID) of plugin (eg "sine_fcaa")
 * \arg \b poly (integer-boolean) - Whether node is polyphonic (0 = false, 1 = true)
 *
 * \li This is only here to provide backwards compatibility for old patches that store plugin
 * references (particularly LADSPA) as libname, label.
 * </p> \n \n
 */
int
OSCReceiver::m_create_node_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* node_path   = &argv[1]->s;
	const char* type        = &argv[2]->s;
	const char* lib_name    = &argv[3]->s;
	const char* plug_label  = &argv[4]->s;
	const int   poly        =  argv[5]->i;
	
	// FIXME Event-ize
	
	Plugin* plugin = new Plugin();
	plugin->set_type(type);
	plugin->lib_name(lib_name);
	plugin->plug_label(plug_label);

	if (poly != 0 && poly != 1) {
		OSCResponder(addr, response_id).respond_error("Invalid poly parameter in create_node");
		return 0;
	}

	AddNodeEvent* ev = new AddNodeEvent(new OSCResponder(addr, response_id), node_path, plugin, (poly == 1));
	om->pre_processor()->push(ev);

	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/destroy - Removes (destroys) a Patch or a Node
 * \arg \b response-id (integer)
 * \arg \b node-path (string) - Full path of the object </p> \n \n
 */
int
OSCReceiver::m_destroy_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* node_path   = &argv[1]->s;
	
	DestroyEvent* ev = new DestroyEvent(new OSCResponder(addr, response_id), node_path);
	om->pre_processor()->push(ev);
		
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/connect - Connects two ports (must be in the same patch)
 * \arg \b response-id (integer)
 * \arg \b src-port-path (string) - Full path of source port
 * \arg \b dst-port-path (string) - Full path of destination port </p> \n \n
 */
int
OSCReceiver::m_connect_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr          = lo_message_get_source(msg);
	const int   response_id   =  argv[0]->i;
	const char* src_port_path = &argv[1]->s;
	const char* dst_port_path = &argv[2]->s;

	ConnectionEvent* ev = new ConnectionEvent(new OSCResponder(addr, response_id), src_port_path, dst_port_path);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/disconnect - Disconnects two ports.
 * \arg \b response-id (integer)
 * \arg \b src-port-path (string) - Full path of source port
 * \arg \b dst-port-path (string) - Full path of destination port </p> \n \n
 */
int
OSCReceiver::m_disconnect_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr          = lo_message_get_source(msg);
	const int   response_id   =  argv[0]->i;
	const char* src_port_path = &argv[1]->s;
	const char* dst_port_path = &argv[2]->s;

	DisconnectionEvent* ev = new DisconnectionEvent(new OSCResponder(addr, response_id), src_port_path, dst_port_path);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/disconnect_all - Disconnect all connections to/from a node.
 * \arg \b response-id (integer)
 * \arg \b node-path (string) - Full path of node. </p> \n \n
 */
int
OSCReceiver::m_disconnect_all_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* node_path   = &argv[1]->s;

	DisconnectNodeEvent* ev = new DisconnectNodeEvent(new OSCResponder(addr, response_id), node_path);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/set_port_value - Sets the value of a port for all voices (both AR and CR)
 * \arg \b response-id (integer)
 * \arg \b port-path (string) - Name of port
 * \arg \b value (float) - Value to set port to
 */
int
OSCReceiver::m_set_port_value_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* port_path   = &argv[1]->s;
	const float value       =  argv[2]->f;

	SetPortValueEvent* ev = new SetPortValueEvent(new OSCResponder(addr, response_id), port_path, value);
	om->event_queue()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/set_port_value - Sets the value of a port for a specific voice (both AR and CR)
 * \arg \b response-id (integer)
 * \arg \b port-path (string) - Name of port
 * \arg \b voice (integer) - Voice to set port value for
 * \arg \b value (float) - Value to set port to
 */
int
OSCReceiver::m_set_port_value_voice_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* port_path   = &argv[1]->s;
	const int   voice       =  argv[2]->i;
	const float value       =  argv[3]->f;

	SetPortValueEvent* ev = new SetPortValueEvent(new OSCResponder(addr, response_id), voice, port_path, value);
	om->event_queue()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/set_port_value_slow - Sets the value of a port for all voices (as a SlowEvent)
 * \arg \b response-id (integer)
 * \arg \b port-path (string) - Name of port
 * \arg \b value (float) - Value to set port to
 *
 * \li This version exists so you can send it immediately after a SlowEvent it may depend on (ie a
 * node creation) and be sure it happens after the event (a normal set_port_value could beat the
 * slow event and arrive out of order). </p> \n \n
 */
int
OSCReceiver::m_set_port_value_slow_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* port_path   = &argv[1]->s;
	const float value       =  argv[2]->f;

	SetPortValueSlowEvent* ev = new SetPortValueSlowEvent(new OSCResponder(addr, response_id), port_path, value);
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/note_on - Triggers a note-on, just as if it came from MIDI
 * \arg \b response-id (integer)
 * \arg \b node-path (string) - Patch of Node to trigger (must be a trigger or note node)
 * \arg \b note-num (int) - MIDI style note number (0-127)
 * \arg \b velocity (int) - MIDI style velocity (0-127)</p> \n \n
 */
int
OSCReceiver::m_note_on_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* node_path   = &argv[1]->s;
	const uchar note_num    =  argv[2]->i;
	const uchar velocity    =  argv[3]->i;

	NoteOnEvent* ev = new NoteOnEvent(new OSCResponder(addr, response_id), node_path, note_num, velocity);
	om->event_queue()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/note_off - Triggers a note-off, just as if it came from MIDI
 * \arg \b response-id (integer)
 * \arg \b node-path (string) - Patch of Node to trigger (must be a trigger or note node)
 * \arg \b note-num (int) - MIDI style note number (0-127)</p> \n \n
 */
int
OSCReceiver::m_note_off_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* patch_path  = &argv[1]->s;
	const uchar note_num    =  argv[2]->i;

	NoteOffEvent* ev = new NoteOffEvent(new OSCResponder(addr, response_id), patch_path, note_num);
	om->event_queue()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/all_notes_off - Triggers a note-off for all voices, just as if it came from MIDI
 * \arg \b response-id (integer)
 * \arg \b patch-path (string) - Patch of patch to send event to </p> \n \n
 */
int
OSCReceiver::m_all_notes_off_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* patch_path  = &argv[1]->s;
	
	AllNotesOffEvent* ev = new AllNotesOffEvent(new OSCResponder(addr, response_id), patch_path);
	om->event_queue()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/synth/midi_learn - Initiate MIDI learn for a given (MIDI Control) Node
 * \arg \b response-id (integer)
 * \arg \b node-path (string) - Patch of the Node that should learn the next MIDI event.
 * 
 * \li This of course will only do anything for MIDI control nodes.  The node will learn the next MIDI
 * event that arrives at it's MIDI input port - no behind the scenes voodoo happens here.  It is planned
 * that a plugin specification supporting arbitrary OSC commands for plugins will exist one day, and this
 * method will go away completely. </p> \n \n
 */
int
OSCReceiver::m_midi_learn_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;
	const char* patch_path  = &argv[1]->s;
	
	MidiLearnEvent* ev = new MidiLearnEvent(new OSCResponder(addr, response_id), patch_path);
	om->pre_processor()->push(ev);
	
	return 0;
}


#ifdef HAVE_LASH
/** \page engine_osc_namespace
 * <p> \b /om/lash/restore_done - Notify LASH restoring is finished and connections can be made.
 * \arg \b response-id (integer)
 */
int
OSCReceiver::m_lash_restore_done_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	const int   response_id =  argv[0]->i;

	LashRestoreDoneEvent* ev = new LashRestoreDoneEvent(new OSCResponder(addr, response_id));
	om->pre_processor()->push(ev);

	return 0;
}
#endif // HAVE_LASH


/** \page engine_osc_namespace
 * <p> \b /om/metadata/set - Sets a piece of metadata, associated with a synth-space object (node, etc)
 * \arg \b response-id (integer)
 * \arg \b object-path (string) - Full path of object to associate metadata with
 * \arg \b key (string) - Key (index) for new piece of metadata
 * \arg \b value (string) - Value of new piece of metadata
 */
int
OSCReceiver::m_metadata_set_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	int         response_id =  argv[0]->i;
	const char* node_path   = &argv[1]->s;
	const char* key         = &argv[2]->s;
	const char* value       = &argv[3]->s;
	
	SetMetadataEvent* ev = new SetMetadataEvent(new OSCResponder(addr, response_id),
		node_path, key, value);
	
	om->pre_processor()->push(ev);

	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/metadata/responder - Requests the engine send a piece of metadata, associated with a synth-space object (node, etc)
 * \arg \b response-id (integer)
 * \arg \b object-path (string) - Full path of object metadata is associated with
 * \arg \b key (string) - Key (index) for piece of metadata
 *
 * \li Reply will be sent to client registered with the source address of this message.</p> \n \n
 */
int
OSCReceiver::m_metadata_get_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	int         response_id =  argv[0]->i;
	const char* node_path   = &argv[1]->s;
	const char* key         = &argv[2]->s;
	
	RequestMetadataEvent* ev = new RequestMetadataEvent(new OSCResponder(addr, response_id),
		node_path, key);

	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/responder/plugins - Requests the engine send a list of all known plugins.
 * \arg \b response-id (integer)
 *
 * \li Reply will be sent to client registered with the source address of this message.</p> \n \n
 */
int
OSCReceiver::m_request_plugins_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	int         response_id =  argv[0]->i;
	
	RequestPluginsEvent* ev = new RequestPluginsEvent(new OSCResponder(addr, response_id));
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/responder/all_objects - Requests the engine send information about \em all objects (patches, nodes, etc)
 * \arg \b response-id (integer)
 * 
 * \li Reply will be sent to client registered with the source address of this message.</p> \n \n
 */
int
OSCReceiver::m_request_all_objects_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	int         response_id =  argv[0]->i;

	RequestAllObjectsEvent* ev = new RequestAllObjectsEvent(new OSCResponder(addr, response_id));
	om->pre_processor()->push(ev);
	
	return 0;
}


/** \page engine_osc_namespace
 * <p> \b /om/responder/port_value - Requests the engine send the value of a port.
 * \arg \b response-id (integer)
 * \arg \b port-path (string) - Full path of port to send the value of </p> \n \n
 *
 * \li Reply will be sent to client registered with the source address of this message.</p> \n \n
 */
int
OSCReceiver::m_request_port_value_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	lo_address  addr        = lo_message_get_source(msg);
	int         response_id =  argv[0]->i;
	const char* port_path   = &argv[1]->s;

	RequestPortValueEvent* ev = new RequestPortValueEvent(new OSCResponder(addr, response_id), port_path);
	om->pre_processor()->push(ev);

	return 0;
}


#ifdef HAVE_DSSI
int
OSCReceiver::m_dssi_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg)
{
	string node_path(path);

	if (node_path.substr(0, 5) != "/dssi")
		return 1;
	
	string command = node_path.substr(node_path.find_last_of("/")+1);
	node_path = node_path.substr(5); // chop off leading "/dssi/"
	node_path = node_path.substr(0, node_path.find_last_of("/")); // chop off command at end
	
	//cout << "DSSI:  Got message " << command << " for node " << node_path << endl;

	SlowEvent* ev = NULL;
	
	if (command == "update" && !strcmp(types, "s"))
		ev = new DSSIUpdateEvent(NULL, node_path, &argv[0]->s);
	else if (command == "control" && !strcmp(types, "if"))
		ev = new DSSIControlEvent(NULL, node_path, argv[0]->i, argv[1]->f);
	else if (command == "configure" && ~strcmp(types, "ss"))
		ev = new DSSIConfigureEvent(NULL, node_path, &argv[0]->s, &argv[1]->s);
	else if (command == "program" && ~strcmp(types, "ii"))
		ev = new DSSIProgramEvent(NULL, node_path, argv[0]->i, argv[1]->i);

	if (ev != NULL)
		om->pre_processor()->push(ev);
	else
		cerr << "[OSCReceiver] Unknown DSSI command received: " << path << endl;

	return 0;
}
#endif


//  Static Callbacks //


// Display incoming OSC messages (for debugging purposes)
int
OSCReceiver::generic_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg, void* user_data)
{
    printf("[OSCMsg] %s\n", path);
    
	for (int i=0; i < argc; ++i) {
		printf("         '%c'  ", types[i]);
		lo_arg_pp(lo_type(types[i]), argv[i]);
		printf("\n");
    }
    printf("\n");

	return 1;  // not handled
}


int
OSCReceiver::unknown_cb(const char* path, const char* types, lo_arg** argv, int argc, lo_message msg, void* user_data)
{
	cerr << "Unknown command " << path << " (" << types << "), sending error.\n";

	string error_msg = "Unknown command: ";
	error_msg.append(path).append(" ").append(types);

	om->osc_sender()->send_error(error_msg);

	return 0;
}


} // namespace Om
